#!/usr/bin/python3

from setuptools import setup

setup(name='ubuntucoreupgrader',
      packages=['ubuntucoreupgrader'],
      test_suite="ubuntucoreupgrader.tests",
      scripts=['bin/ubuntu-core-upgrade'],
      setup_requires=[
          "flake8",
      ],
      )
